#ifndef __DWC_LIB_H
#define __DWC_LIB_H

/*--------------------------------------------------------------------------
 Library for building clients by Dr.Web daemon
                                        ID Antivirus Lab (http://www.drweb.ru)
                                        http://www.drweb.ru/linux-bsd
                                        http://www.drweb.ru/linux-bsd/forum
--------------------------------------------------------------------------*/

#include <sys/types.h>
#include "dwc_scan.h"

/* -- RETURN CODES ----------------------------------------------------- */
#define DWL_NOERROR                 (0)   /* no errors */
#define DWL_COMMONERROR             (-1)  /* returned by system */
#define DWL_IPCSYSTEM               (-2)  /* in internal transport system */  
#define DWL_BADFILE                 (-3)  /* on file passed to lib */
#define DWL_BADPARAMS               (-4)  /* incorrect values of parameters */
#define DWL_NOTSUPPORTED            (-5)  /* this option not supported */
#define DWL_ALREADYEXEC             (-6)  /* try make operation that can be performed only once */
#define DWL_RECVERROR               (-7)  /* on recv from daemon */
#define DWL_SENDERROR               (-8)  /* on send to daemon */
#define DWL_NOMEMORY                (-9)  /* no memory for internal use */
#define DWL_NOTINIT                 (-10) /* library was not inited */

/* -- Transport types ------------------------------------------------- */
/* Values for Type in startupLib */
#define DWL_TYPE_TCP_SOCKET         (1) /* use TCP sockets to connect with daemon */
#define DWL_TYPE_UNIX_SOCKET        (2) /* use UNIX sockets to connect with daemon - under Windows not supported */
#define DWL_TYPE_READPIDFILE        (3) /* read daemon pid-file to gets transport properties - under Windows not supported */
#define DWL_TYPE_DEFAULT            (4) /* use predefined transport - currently TCP */

/* -- Value of parameter FileType for scanFile ------------------------- */
#define DWL_FT_NORMALFILE           (1) /* read file and send him thru transport to daemon for scan */
#define DWL_FT_DISKFILE             (2) /* send to daemon only filename, implemented only on platform with existen daemon version */

/* -- PROTOTYPES ------------------------------------------------------ */
/* Enable/Disable output : Can (should?) be used before startup
 * IN: Quiet - enable (!0) or disable (0) all output (work and debug)
 * IN: Debug - enable (!0) or disable (0) debug output
 * IN: Syslog - output to syslog(!0) or to console(0)
 * DEFAULT OUTPUT MODE: all disabled
 */
extern void outputLib( int Quiet, int Debug, int Syslog );

/* Output information, according to output settings 
 * w - work output, d - debug output
 * IN: fmt - format string (see doc to printf)
 */ 
extern void wprintLib( char const* fmt, ... );
extern void dprintLib( char const* fmt, ... );

/* Return error message appropriate for library return codes, never NULL */
extern const char* getLibError( int code );

/* Return error message appropriate for daemon return codes, never NULL */
extern const char* getDaemonCode( int code );

/* Initialize library : made it  before use any belows functions.
 * IN: Timeout - time in microseconds for network operations;
 * IN: Type - one of defined Transport types : 
 *         DWL_TYPE_DEFAULT : no additional parameters - use all by default (Timeout ignored)
 *         DWL_TYPE_TCP_SOCKET : 
 * IN:         const char*  HostName - dot-name or domain name of host with runnig daemon
 * IN:         unsigned int Port - tcp-port number (in host order)
 *         DWL_TYPE_UNIX_SOCKET :
 * IN:         const char*  SocketFileName, name of unix-domain socket file
 *         DWL_TYPE_READPIDFILE :
 * IN:         const char*  PidFileName, name of daemon pid-file
 *         Note: 
 *           1) all *Name can be NULL or begins with NULLCHAR('\0') - it's mean that will be used default names
 *           2) Port can be 0 - it's mean use default port
 * Return: codes see before
 */
extern int  startupLib( int Type, int Timeout, ... );

/* Deinitialize library - do it before exit or startup with new parameters
 * Return: codes see before
 */
extern int  cleanupLib( void );

/* Show daemon ID string
 * IN/OUT: log - ptr to buffer or NULL if want use internal logging 
 * IN/OUT: loglen - len of buffer as IN, and len of OUT log-string
 * Return: codes see before
 */
extern int  showID( char* log, size_t* loglen );

/* Show daemon version 
 * IN/OUT: log - ptr to buffer or NULL if want use internal logging 
 * IN/OUT: loglen - len of buffer as IN, and len of OUT log-string
 * Return: codes see before
 */
extern int  showVersion( char* log, size_t* loglen );

/* Show bases loaded by daemon 
 * IN/OUT: log - ptr to buffer or NULL if want use internal logging
 * IN/OUT: loglen - len of buffer as IN, and len of OUT log-string
 * Return: codes see before
 */
extern int  showBases( char* log, size_t* loglen );

/* Scan memory buffer :
 * IN: Buffer - pointer to memory for scan
 * IN: BufferLen - size of buffer
 * IN: ScanOptions - set of options from <dwc_scan.h>     
 * OUT: result_ptr - code returned by daemon
 * IN/OUT: log - ptr to buffer or NULL if want use internal logging
 * IN/OUT: loglen - len of buffer as IN, and len of OUT log-string
 * Return: codes see before
 */
extern int  scanBuffer( void* Buffer, size_t BufferLen, int ScanOptions, int* result_ptr, char* log, size_t* loglen );

/* Scan file :
 * IN: Path - pointer to path name (or filename for normal files)
 * IN: PathLen - length of path name (library dont checks for correct this value)
 * IN: FileType - one of
 *         DWL_FT_NORMALFILE - open, read file and send to daemon 
 *         DWL_FT_DISKFILE - send to daemon only filename and daemon try verify it as local file
 * IN: ScanOptions - set of options from <dwc_scan.h>     
 * OUT: result_ptr - code returned by daemon
 * IN/OUT: log - ptr to buffer or NULL if want use internal logging
 * IN/OUT: loglen - len of buffer as IN, and len of OUT log-string
 * Return: codes see before
 */
extern int  scanFile( char* Path, size_t PathLen, int FileType, int ScanOptions, int* result_ptr, char* log, size_t* loglen );

/* Scan part file (useful for checks mailboxes - not implemented under not *nix platfroms) :
 * IN: Path - pointer to path name (seen by daemon) to file
 * IN: PathLen - length of path name (library dont checks for correct this value)
 * IN: Offset - offset from head of file to scan
 * IN: Length - length of part file to scan
 * IN: ScanOptions - set of options from <dwc_scan.h>     
 * OUT: result_ptr - code returned by daemon
 * IN/OUT: log - ptr to buffer or NULL if want use internal logging 
 * IN/OUT: loglen - len of buffer as IN, and len of OUT log-string
 * Return: codes see before
 */
extern int  scanFilePart( char* Path, size_t PathLen, size_t Offset, size_t Length, int ScanOptions, int* result_ptr, char* log, size_t* loglen );

#endif /* __DWC_LIB_H */
