#ifndef __DWC_TIMO_C
#define __DWC_TIMO_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_timo.c,v 1.14 2001/06/15 03:58:37 cvser Exp $";
#endif

#include <errno.h>

#if defined(__UNIX_DW)
#include <sys/time.h>
#if defined(__SIGNAL_BASED_DW) && defined(__THREAD_BASED_DW)
#error "You can not defined __SIGNAL_BASED_DW && __THREAD_BASED_DW together"
#endif /* __SIGNAL_BASED_DW && __THREAD_BASED_DW */
#if !defined(__SIGNAL_BASED_DW) && !defined(__THREAD_BASED_DW)
#error "You should define __SIGNAL_BASED_DW or __THREAD_BASED_DW"
#endif /* not __SIGNAL_BASED_DW && not __THREAD_BASED_DW */
#if defined(__SIGNAL_BASED_DW)
#include <signal.h>
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
#include <pthread.h>
#include <sched.h>
#endif /* __THREAD_BASED_DW */
#include <string.h> 
#endif /* __UNIX_DW */

#if defined(__WIN32_DW)
#define __THREAD_BASED_DW
#include <windows.h>
#endif

#include "dwc_timo.h"
#include "dwc_lib.h"
#include "dwc_type.h"
#include "dwc_err.h"
#include "dwc_macro.h"
#include "bsd_queue.h"

typedef struct _DWC_timer_t {
    TAILQ_ENTRY( _DWC_timer_t )     all;
    TAILQ_ENTRY( _DWC_timer_t )     active;
    DWC_bool_t                      alarmed;
    DWC_bool_t                      waited;
    DWC_bool_t                      deletable;
    DWC_timeout_t                   value;
    DWC_TimeoutProc_t               proc;
    DWC_pointer_t                   data;
} DWC_timer_t;

typedef TAILQ_HEAD( _DWC_tqhead_t, _DWC_timer_t ) DWC_tqhead_t;

typedef struct _DWC_tqueue_t {
    DWC_tqhead_t                    all_head;
    DWC_tqhead_t                    active_head;
    DWC_bool_t                      inited;
    DWC_bool_t                      locked;
    DWC_timeout_t                   slice;
#if defined(__UNIX_DW)
#if defined(__SIGNAL_BASED_DW)    
    struct sigaction                oldact;
    sigset_t                        sigset;
    DWC_bool_t                      in_signal;
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    pthread_t                       thread;
    pthread_mutex_t                 mutex;
    pthread_t                       locker;
    pthread_cond_t                  cond;
    pthread_mutex_t                 cond_mutex;
#endif /* __THREAD_BASED_DW */    
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    HANDLE                          NewValue;
    HANDLE                          ExitThread;
    HANDLE                          Mutex;
    HANDLE                          Thread;
    DWORD                           ThreadID;
#endif /* __WIN32_DW */
} DWC_tqueue_t;

static DWC_tqueue_t      TimerQueue = {
    TAILQ_HEAD_INITIALIZER( TimerQueue.all_head ),
    TAILQ_HEAD_INITIALIZER( TimerQueue.active_head ),
    FALSE,
    FALSE,
    0    /* others components are system-dependent */
};

#if defined(__UNIX_DW)

#if defined(__SIGNAL_BASED_DW)

#define TIMER_NUM_SLICES    (16)
#if defined(__Linux__)
#define TIMER_MIN_SLICE     (10000)
#else /* FreeBSD. Default for other - not tested */
#define TIMER_MIN_SLICE     (20000)
#endif /* Linux */

static void                 UnixSigAlarm ( int signo );

#endif /* __SIGNAL_BASED_DW */

#if defined(__THREAD_BASED_DW)

#define TIMER_NUM_SLICES    (1)
#define TIMER_MIN_SLICE     (20000)

static void*                TimerThread ( void* arg );

#endif /* __THREAD_BASED_DW */
            
#endif /* __UNIX_DW */

#if defined(__WIN32_DW)

#define TIMER_NUM_SLICES    (2)
#define TIMER_MIN_SLICE     (55000) /* 55 msec for Win95 better */

static DWORD WINAPI         TimerThread ( LPVOID param );

#endif /* __WIN32_DW */
            
static DWC_bool_t           SelfLocking( void );
static DWC_bool_t           LockQueue ( DWC_tqueue_t *queue );
static DWC_bool_t           UnLockQueue ( DWC_tqueue_t *queue );
static DWC_error_code_t     RunIntTimer( DWC_timeout_t value );
static DWC_error_code_t     StopIntTimer( void );
static void                 ReceiveTick( void );
static void                 AlarmTimer( DWC_timer_hnd_t Timer );
static void                 RemoveTimer( DWC_timer_hnd_t Timer, DWC_bool_t Finally );

#if defined(__UNIX_DW)
/* --------------------------------------------------------------------- */
#if defined(__SIGNAL_BASED_DW)
static void UnixSigAlarm ( int signo )
{
    if( signo == SIGALRM )
    {
        TimerQueue.in_signal = TRUE;
        ReceiveTick();
        TimerQueue.in_signal = FALSE;
    }
}
#endif /* __SIGNAL_BASED_DW */
/* --------------------------------------------------------------------- */
#if defined(__THREAD_BASED_DW)
static void* TimerThread ( void* arg )
{
    int               retcode;
    DWC_bool_t        TickNow, NeedExit;
    DWC_timeout_t     oslice;
    struct timeval    interval, base, result;
    struct timespec   timeout;
    
    NeedExit = FALSE;
    if( pthread_mutex_lock( &TimerQueue.cond_mutex ) != 0 )
    {
        dprintLib( "can't lock cond mutex - %s.\n", strerror( errno ) );
        pthread_exit( NULL );        
    }
    while( !NeedExit )
    {
        TickNow = FALSE;
        if( TimerQueue.slice != 0 )
        {
            interval.tv_sec = TimerQueue.slice / 1000000;
            interval.tv_usec = TimerQueue.slice % 1000000;
            gettimeofday( &base, NULL );
            timeradd( &interval, &base, &result );
            TIMEVAL_TO_TIMESPEC( &result, &timeout );
            oslice = TimerQueue.slice;
          again:
            retcode = pthread_cond_timedwait( &TimerQueue.cond, &TimerQueue.cond_mutex, &timeout );
            if( retcode == ETIMEDOUT )
            {
                TickNow = TRUE;
                retcode = 0;
            }
            else if( retcode == EINTR )
            {
                goto again;
            }
        }
        else
        {
            retcode = pthread_cond_wait( &TimerQueue.cond, &TimerQueue.cond_mutex );
        }
        if( retcode != 0 )
        {   /* error */
            NeedExit = TRUE;
        }
        if( TickNow )
        {
            ReceiveTick();
        }
    }
    pthread_exit( NULL );
}
#endif /* __THREAD_BASED_DW */
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
/* --------------------------------------------------------------------- */
static DWORD WINAPI TimerThread( LPVOID ignore )
{
    DWORD      timeout,retcode;
    BOOL       NeedExit = FALSE;
    HANDLE     handles[] = {
    /* 0 */  TimerQueue.ExitThread,
    /* 1 */  TimerQueue.NewValue,
    };

    while( !NeedExit )
    {
        if( TimerQueue.slice != 0 )
        {
            timeout = TimerQueue.slice / 1000; /* slice in microsecs and timeout in millisecs */
        }
        else
        {
            timeout = INFINITE;
        }
        retcode = WaitForMultipleObjects( 2, handles, FALSE, timeout );
        if( retcode == WAIT_TIMEOUT )
        {
            /* timeouted */
            ReceiveTick();
        }
        else if( retcode == WAIT_OBJECT_0 )
        {
            /* ExitThread event */
            NeedExit = TRUE;
            retcode = 0;
        }
        else if( retcode == WAIT_OBJECT_0 + 1 )
        {
            /* Set New Timeout value */
            if( !ResetEvent( TimerQueue.NewValue ) )
            {
                dprintLib( "Timer thread: can't reset event.\n" );
                NeedExit = TRUE;
            }
        }
        else 
        {
            /* any error */
            NeedExit = TRUE;
            retcode = 1;
        }
    }
    return retcode;
}
#endif /* __WIN32_DW */

/* --------------------------------------------------------------------- */
static DWC_bool_t LockQueue ( DWC_tqueue_t *queue )
{
#if defined(__UNIX_DW)    
#if defined(__SIGNAL_BASED_DW)
    if( !queue->in_signal )
    {
        if( sigprocmask( SIG_BLOCK, &TimerQueue.sigset, NULL ) < 0 )
        {
            return FALSE;
        }
        queue->locked = TRUE;
    }
    return TRUE;
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    int retval;

    retval = pthread_mutex_lock( &queue->mutex );
    if( retval != 0 )
    {
        dprintLib( "can't lock queue mutex - %s.\n", strerror( retval ) );
        return FALSE;
    }
    queue->locked = TRUE;
    queue->locker = pthread_self();
    return TRUE;
#endif /* __THREAD_BASED_DW */
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)    
    DWORD   retval;

    retval = WaitForSingleObject( queue->Mutex, INFINITE );
    if( retval != WAIT_OBJECT_0 )
    {
        dprintLib( "can't lock queue mutex - %d.\n", retval );
        return FALSE;
    }
    queue->locked = TRUE;
    return TRUE;
#endif /* __WIN32_DW */    
}

/* --------------------------------------------------------------------- */
static DWC_bool_t UnLockQueue ( DWC_tqueue_t *queue )
{
#if defined(__UNIX_DW)    
#if defined(__SIGNAL_BASED_DW)
    if( !queue->in_signal )
    {
        queue->locked = FALSE;
        if( sigprocmask( SIG_UNBLOCK, &TimerQueue.sigset, NULL ) < 0 )
        {
            return FALSE;
        }
    }
    return TRUE;
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    int retval;

    queue->locked = FALSE;
    retval = pthread_mutex_unlock( &queue->mutex );
    if( retval != 0 )
    {
        queue->locked = TRUE;
        dprintLib( "can't unlock queue mutex - %s.\n", strerror( retval ) );
        return FALSE;
    }
    return TRUE;
#endif /* __THREAD_BASED_DW */
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)    
    queue->locked = FALSE;
    if( !ReleaseMutex( queue->Mutex ) )
    {
        queue->locked = TRUE;
        dprintLib( "can't unlock queue mutex - %d.\n", errno );
        return FALSE;
    }
    return TRUE;
#endif /* __WIN32_DW */    
}

/* --------------------------------------------------------------------- */
static DWC_error_code_t RunIntTimer( DWC_timeout_t timeout )
{
#if defined(__UNIX_DW) && defined(__SIGNAL_BASED_DW)    
    struct itimerval  itv;
#endif /* __SIGNAL_BASED_DW && __UNIX_DW */

    if( timeout == 0 )
    {
        TimerQueue.slice = 0;
    }
    else if( timeout < TIMER_MIN_SLICE * TIMER_NUM_SLICES )
    {
        TimerQueue.slice = TIMER_MIN_SLICE;
    }    
    else
    {
        TimerQueue.slice = timeout / TIMER_NUM_SLICES;
    }
    /* all systems has fixed quantized step of timers */
    TimerQueue.slice -= TimerQueue.slice % TIMER_MIN_SLICE;
#if defined(__UNIX_DW)    
#if defined(__SIGNAL_BASED_DW)    
    itv.it_value.tv_sec = TimerQueue.slice / 1000000;
    itv.it_value.tv_usec = TimerQueue.slice % 1000000;
    itv.it_interval = itv.it_value;
    if( setitimer( ITIMER_REAL, &itv, NULL ) < 0 )
    {
        dprintLib( "Can't setitimer().\n" );
        return DWC_NORESOURCE;
    }
    return DWC_NOERROR;    
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    if( pthread_cond_signal( &TimerQueue.cond ) != 0 )
    {
        dprintLib( "Can't singal condition to timer thread.\n" );
        return DWC_NORESOURCE;
    }
    return DWC_NOERROR;    
#endif /* __THREAD_BASED_DW */
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( !SetEvent( TimerQueue.NewValue ) )
    {         
        dprintLib( "Can't set event for timer thread.\n" );
        return DWC_COMMONERROR;
    }
    return DWC_NOERROR;
#endif /* __WIN32_DW */    
}

/* --------------------------------------------------------------------- */
static DWC_error_code_t StopIntTimer( void )
{
    return RunIntTimer( 0 );
}

/* --------------------------------------------------------------------- */
static DWC_bool_t SelfLocking( void )
{
#if defined(__UNIX_DW)
#if defined(__SIGNAL_BASED_DW)
    return TimerQueue.in_signal;
#endif /* __SIGNAL_BASED_DW */
        /* ASV: it looks unsafe, but it should work, because
           if queue was locked before next check to equal
           it can be unlocked and locked again, BUT not this thread (we
           here) end checks fails. If we have mutex (and set locked),
           noboby can't unlock and between two checks state of locked
           can't be changed */
#if defined(__THREAD_BASED_DW)
    if( TimerQueue.locked
        &&
        (pthread_equal( TimerQueue.locker, pthread_self() ) != 0) )
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
#endif /* __THREAD_BASED_DW */    
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( TimerQueue.locked 
        &&
        (TimerQueue.ThreadID == GetCurrentThreadId()) )
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
#endif /* __WIN32_DW */
}

/* --------------------------------------------------------------------- */
static void ReceiveTick( void )
{
    DWC_timer_hnd_t   Timer;
    DWC_timeout_t     timo;

    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Info: timer handler can't lock engine - lose slice.\n" );
        return;
    }
    timo = TimerQueue.slice;
  cont_l:
    Timer = TAILQ_FIRST( &TimerQueue.active_head );
    if( !Timer )
    {    
        StopIntTimer();
    }
    else if( timo < Timer->value )
    {
        Timer->value -= timo;       
    }
    else 
    {
        timo -= Timer->value;        
        AlarmTimer( Timer );
        goto cont_l;
    }
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Error: can't unlock engine.\n" );
    }
}

/* --------------------------------------------------------------------- */
static void AlarmTimer( DWC_timer_hnd_t Timer )
{
    if( !Timer->alarmed )
    {        
        Timer->alarmed = TRUE;
        if( Timer->proc )
        {
            Timer->proc( Timer->data );
        }
        RemoveTimer( Timer, Timer->deletable );
    }
    else
    {
        dprintLib( "Warning: timer already alarmed.\n" );
    }
}

/* --------------------------------------------------------------------- */
static void RemoveTimer( DWC_timer_hnd_t Timer, DWC_bool_t Finally )
{
    if( !Timer->waited )
    {
        Timer->waited = TRUE;
        TAILQ_REMOVE( &TimerQueue.active_head, Timer, active );        
        if( TAILQ_EMPTY( &TimerQueue.active_head ) )
        {
            if( StopIntTimer() != DWC_NOERROR )
            {
                dprintLib( "Error: Can't stop internal timer.\n" );
            }
        }
        else
        {
            if( RunIntTimer( TAILQ_FIRST( &TimerQueue.active_head )->value ) != DWC_NOERROR )
            {
                dprintLib( "Error: Can't set internal timer.\n" );
            }
        }
    }
    if( Finally )
    {    
        TAILQ_REMOVE( &TimerQueue.all_head, Timer, all );
        DWC_Free( Timer );
    }
}

/* --------------------------------------------------------------------- */
DWC_error_code_t InitTimerEngine( void )
{
#if defined(__UNIX_DW)
#if defined(__SIGNAL_BASED_DW)    
    struct sigaction  sig;
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)
    int                retcode;
#endif /* __THREAD_BASED_DW */    
#endif/* __UNIX_DW */
    
    if( TimerQueue.inited )
    {
        dprintLib( "Timer engine already inited.\n" );
        return DWC_ALREADYPERFORM;
    }
    TimerQueue.locked = TRUE;
    TimerQueue.slice = 0;
#if defined(__UNIX_DW)
#if defined(__SIGNAL_BASED_DW)    
    if( StopIntTimer() != DWC_NOERROR )
    {
        dprintLib( "Can't set timer.\n" );
        dprintLib( "Error(%u): %s \n", errno, strerror(errno) );
        return DWC_NORESOURCE;
    }
    sig.sa_handler = UnixSigAlarm;
    sig.sa_flags = SA_RESTART;
    sigemptyset( &sig.sa_mask );
    if( sigaction( SIGALRM, &sig, &TimerQueue.oldact ) < 0 )
    {
        dprintLib( "Can't set signal handler for SIGALRM.\n" );
        dprintLib( "Error(%u): %s \n", errno, strerror(errno) );
        return DWC_COMMONERROR;
    }
    sigemptyset( &TimerQueue.sigset );
    sigaddset( &TimerQueue.sigset, SIGALRM );    
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)    
    /* use default 'fast' mutex */
    pthread_mutex_init( &TimerQueue.mutex, NULL );
    pthread_mutex_init( &TimerQueue.cond_mutex, NULL );
    retcode = pthread_cond_init( &TimerQueue.cond, NULL );
    if( retcode != 0 )
    {
        dprintLib( "Can't create condition variable - %s.\n", strerror(retcode) );
        if( pthread_mutex_destroy( &TimerQueue.mutex ) != 0 )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        if( pthread_mutex_destroy( &TimerQueue.cond_mutex ) != 0 )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        return DWC_NORESOURCE;
    }
    retcode = pthread_create( &TimerQueue.thread, NULL, TimerThread, NULL );
    if( retcode != 0 )
    {
        dprintLib( "Can't create timer thread - %s.\n", strerror(retcode) );
        if( pthread_cond_destroy( &TimerQueue.cond ) != 0 )
        {
            dprintLib( "Can't destroy condition variable.\n" );
        }
        if( pthread_mutex_destroy( &TimerQueue.mutex ) != 0 )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        if( pthread_mutex_destroy( &TimerQueue.cond_mutex ) != 0 )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        return DWC_NORESOURCE;
    }
#endif /* __THREAD_BASED_DW */
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    TimerQueue.Mutex = CreateMutex( NULL, FALSE, "MutexForLockTimerEngine" );
    if( !TimerQueue.Mutex )
    {
        dprintLib( "Can't create mutex for locking queue.\n" );
        return DWC_NORESOURCE;
    }
    TimerQueue.NewValue = CreateEvent( NULL, TRUE, FALSE, "SetNewValueForTimeoutThread" );
    if( !TimerQueue.NewValue )
    {
        dprintLib( "Can't create event for set new values to timer thread.\n" );
        if( !CloseHandle( TimerQueue.Mutex ) )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        return DWC_NORESOURCE;
    }
    TimerQueue.ExitThread = CreateEvent( NULL, TRUE, FALSE, "ExitForTimeoutThread" );
    if( !TimerQueue.ExitThread )
    {
        dprintLib( "Can't create event for stops timer thread.\n" );
        if( !CloseHandle( TimerQueue.Mutex ) )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        if( !CloseHandle( TimerQueue.NewValue ) )
        {
            dprintLib( "Can't destroy event for set value.\n" );
        }
        return DWC_NORESOURCE;
    }
    TimerQueue.Thread = CreateThread( NULL, 4096, TimerThread, NULL, CREATE_SUSPENDED, &TimerQueue.ThreadID );
    if( !TimerQueue.Thread )
    {
        dprintLib( "Can't create timer thread.\n" );
        if( !CloseHandle( TimerQueue.Mutex ) )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        if( !CloseHandle( TimerQueue.NewValue ) )
        {
            dprintLib( "Can't destroy event for set value.\n" );
        }
        if( !CloseHandle( TimerQueue.ExitThread ) )
        {
            dprintLib( "Can't destroy event for exiting thread.\n" );
        }
        return DWC_NORESOURCE;
    }
    /* if next call fails it's not critical - just run as std thread */
    SetThreadPriority( TimerQueue.Thread, THREAD_PRIORITY_TIME_CRITICAL );
#endif /* __WIN32_DW */
    TimerQueue.locked = FALSE;
    TimerQueue.inited = TRUE; 
#if defined(__WIN32_DW)    
    if( ResumeThread( TimerQueue.Thread ) == -1 )
    {
        dprintLib( "Can't run timer thread.\n" );
        TimerQueue.inited = FALSE; 
        if( !TerminateThread( TimerQueue.Thread, 0 ) )
        {
            dprintLib( "Can't terminate timer thread.\n" );
        }
        if( !CloseHandle( TimerQueue.Mutex ) )
        {
            dprintLib( "Can't destroy mutex.\n" );
        }
        if( !CloseHandle( TimerQueue.NewValue ) )
        {
            dprintLib( "Can't destroy event for set value.\n" );
        }
        if( !CloseHandle( TimerQueue.ExitThread ) )
        {
            dprintLib( "Can't destroy event for exiting thread.\n" );
        }
        return DWC_NORESOURCE;
    }
#endif /* __WIN32_DW */
    return DWC_NOERROR;    
}

/* --------------------------------------------------------------------- */
DWC_error_code_t DeInitTimerEngine( void )
{
    DWC_bool_t        error = FALSE;
    
    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine wasn't inited.\n" );
        return DWC_NOTREADY;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call user_proc() -> and user call us recursively lock()
     */
    if( TimerQueue.locked )
    {
        dprintLib( "Timer engine locked: you shouldn't call Timeout-routines from your timeout handler.\n" );
        return DWC_NOTREADY;
    }
#if defined(__UNIX_DW)
#if defined(__SIGNAL_BASED_DW)    
    if( StopIntTimer() != DWC_NOERROR )
    {
        dprintLib( "Can't stop timer.\n" );
        error = TRUE;
    }
    if( sigaction( SIGALRM, &TimerQueue.oldact, NULL ) < 0 )
    {
        dprintLib( "Can't restore old action for SIGALRM.\n" );
        dprintLib( "Error(%u): %s \n", errno, strerror(errno) );
        error = TRUE;
    }
#endif /* __SIGNAL_BASED_DW */
#if defined(__THREAD_BASED_DW)    
    if( pthread_mutex_lock( &TimerQueue.cond_mutex ) != 0 )
    {
        dprintLib( "Can't lock cond-mutex.\n" );
        error = TRUE;
    }
    if( pthread_cond_signal( &TimerQueue.cond ) != 0 )
    {
        dprintLib( "Can't signal cond var.\n" );
        error = TRUE;
        if( pthread_detach( TimerQueue.thread ) != 0 )
        {
            dprintLib( "Can't detach timer thread.\n" );
        }
    }
    if( pthread_mutex_unlock( &TimerQueue.cond_mutex ) != 0 )
    {
        dprintLib( "Can't unlock cond-mutex.\n" );
        error = TRUE;
    }    
    if( (pthread_cancel( TimerQueue.thread ) != 0)
        ||
        (pthread_join( TimerQueue.thread, NULL ) != 0) )
    {
        dprintLib( "Can't join timer thread.\n" );
        error = TRUE;
    }
    if( pthread_cond_destroy( &TimerQueue.cond ) != 0 )
    {
        dprintLib( "Can't destroy condition variable.\n" );
        error = TRUE;
    }
    if( pthread_mutex_unlock( &TimerQueue.cond_mutex ) != 0 )
    {
        dprintLib( "Can't unlock cond-mutex.\n" );
        error = TRUE;
    }
    if( pthread_mutex_destroy( &TimerQueue.cond_mutex ) != 0 )
    {
        dprintLib( "Can't destroy cond-mutex.\n" );
        error = TRUE;
    }    
#endif /* __THREAD_BASED_DW */    
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    if( !SetEvent( TimerQueue.ExitThread ) )
    {         
        dprintLib( "Can't set event for timer thread.\n" );
        error = TRUE;
    }
    if( WaitForSingleObject( TimerQueue.Thread, INFINITE ) == WAIT_TIMEOUT )
    {
        dprintLib( "Can't exited timer thread.\n" );
        error = TRUE;
    }
    if( !CloseHandle( TimerQueue.ExitThread ) )
    {
        dprintLib( "Can't destroy event for exit.\n" );
        error = TRUE;
    }
    if( !CloseHandle( TimerQueue.NewValue ) )
    {
        dprintLib( "Can't destroy event for set new value.\n" );
        error = TRUE;
    }
#endif /* __WIN32_DW */
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't lock engine to stop.\n" );
        error = TRUE;
    }
    else
    {
        while( !TAILQ_EMPTY( &TimerQueue.all_head ) )
        {            
            RemoveTimer( TAILQ_FIRST( &TimerQueue.all_head), TRUE );
        }        
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Can't unlock engine.\n" );
            error = TRUE;
        }
    }    
#if defined(__UNIX_DW) && defined(__THREAD_BASED_DW)
    if( pthread_mutex_unlock( &TimerQueue.mutex ) != 0 )
    {
        dprintLib( "Can't unlock queue-mutex.\n" );
        error = TRUE;
    }
    if( pthread_mutex_destroy( &TimerQueue.mutex ) != 0 )
    {
        dprintLib( "Can't destroy queue mutex.\n" );
        error = TRUE;
    }
#endif /* __THREAD_BASED_DW && __UNIX_DW */
#if defined(__WIN32_DW)
    if( !CloseHandle( TimerQueue.Mutex ) )
    {
        dprintLib( "Can't destroy queue mutex.\n" );
        error = TRUE;
    }
#endif /* __WIN32_DW */
    return ( !error ? DWC_NOERROR : DWC_COMMONERROR );    
}

/* --------------------------------------------------------------------- */
DWC_timer_hnd_t SetTimeout( DWC_timeout_t timeout, DWC_TimeoutProc_t Proc, DWC_pointer_t Data, DWC_bool_t DelAtAlarm )
{
    DWC_timer_hnd_t   Timer;
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        inserted = FALSE;

    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine wasn't inited.\n" );
        return INVALID_TIMER;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() ) 
    {
        return INVALID_TIMER;
    }    
    if( timeout != INVALID_TIMEOUT )
    {
        if( timeout == INFINITE_TIMEOUT )
        {
            return INFINITE_TIMER;
        }
        Timer = DWC_Malloc( DWC_timer_hnd_t, sizeof(DWC_timer_t) );
        if( !Timer )
        {
            dprintLib( "Can't create timer - no memory.\n" );
            return INVALID_TIMER;
        }
        if( !LockQueue( &TimerQueue ) )
        {
            dprintLib( "Can't lock timer engine on create.\n" );
            DWC_Free( Timer );
            return INVALID_TIMER;
        }
        Timer->alarmed = FALSE;
        Timer->deletable = DelAtAlarm;
        Timer->waited = FALSE;
        Timer->proc = Proc;
        Timer->data = Data;
        if( TAILQ_EMPTY( &TimerQueue.active_head ) )
        {
            TAILQ_INSERT_HEAD( &TimerQueue.active_head, Timer, active) ;
            RunIntTimer( timeout );
        }
        else if( timeout < TAILQ_FIRST(&TimerQueue.active_head)->value )                
        {
            TAILQ_FIRST(&TimerQueue.active_head)->value -= timeout;
            TAILQ_INSERT_HEAD(&TimerQueue.active_head, Timer, active);
            RunIntTimer( timeout );
        }
        else
        {
            TAILQ_FOREACH( tmpTimer, &TimerQueue.active_head, active )
            {
                if( timeout < tmpTimer->value )
                {
                    TAILQ_INSERT_BEFORE( tmpTimer, Timer, active );
                    tmpTimer->value -= timeout;
                    inserted = TRUE;
                    break;
                }
                timeout -= tmpTimer->value;
            }
            if( !inserted )
            {
                TAILQ_INSERT_TAIL( &TimerQueue.active_head, Timer, active );
            }
        }
        Timer->value = timeout;
        TAILQ_INSERT_TAIL( &TimerQueue.all_head, Timer, all );
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Error: can't unlock timer engine.\n" );
            RemoveTimer( Timer, TRUE );
            DWC_Free( Timer );
            return INVALID_TIMER;
        }
        return Timer;
    }
    else
    {
        return INVALID_TIMER;
    }
}

/* --------------------------------------------------------------------- */
DWC_timeout_t RetrieveTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timeout_t     retval = 0;
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        found = FALSE;
    
    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine wasn't inited.\n" );
        return INVALID_TIMEOUT;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return INVALID_TIMEOUT;
    }
    else if( Timer == INFINITE_TIMER )
    {
        return INFINITE_TIMEOUT;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() ) 
    {
        return INVALID_TIMEOUT;
    }    
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't lock timer engine.\n" );
        return INVALID_TIMEOUT;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.active_head, active )
    {
        retval += tmpTimer->value;
        if( tmpTimer == Timer )
        {
            found = TRUE;
            break;
        }
    }
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Error: Can't unlock timer engine.\n" );
        return INVALID_TIMEOUT;
    }
    if( found )
    {
        return retval;
    }
    else
    {
        dprintLib( "Invalid timer: not exist or already alarmed.\n" );
        return INVALID_TIMEOUT;        
    }
}

/* --------------------------------------------------------------------- */
DWC_bool_t AlarmTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        found = FALSE;
    
    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine wasn't inited.\n" );
        return FALSE;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return FALSE;
    }
    else if( Timer == INFINITE_TIMER )
    {
        dprintLib( "Infinite timeout can't be alarmed.\n" );
        return FALSE;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() ) 
    {
        return FALSE;
    }    
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't lock timer engine.\n" );
        return FALSE;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.active_head, active )
    {
        if( tmpTimer == Timer )
        {
            found = TRUE;
            break;
        }
    }
    if( !found )
    {
        dprintLib( "Invalid timer: not exist or already alarmed.\n" );
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Can't unlock timer engine.\n" );
        }
        return FALSE;        
    }
    AlarmTimer( Timer );
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't unlock timer engine.\n" );
        return FALSE;
    }
    return TRUE;
}

/* --------------------------------------------------------------------- */
DWC_bool_t ExpiredTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        retval = FALSE, found = FALSE;
    
    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine wasn't inited.\n" );
        return FALSE;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return FALSE;
    }
    else if( Timer == INFINITE_TIMER )
    {
        return FALSE;
    }
    /* It's guard *ONLY* against one thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() ) 
    {
        return FALSE;
    }    
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't lock timer engine.\n" );
        return FALSE;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.all_head, all )
    {
        if( tmpTimer == Timer )
        {
            found = TRUE;
            retval = Timer->alarmed;
            break;
        }
    }
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't unlock timer engine.\n" );
        return FALSE;
    }
    if( !found )
    {
        dprintLib( "Invalid timer for test: not exist or already deleted.\n" );
        dprintLib( "Requested timer at %p.\n", Timer );
        return FALSE;        
    }
    return retval;
}

/* --------------------------------------------------------------------- */
DWC_bool_t ClearTimeout( DWC_timer_hnd_t Timer )
{
    DWC_timer_hnd_t   tmpTimer;
    DWC_bool_t        found = FALSE;
    
    if( !TimerQueue.inited )
    {
        dprintLib( "Timer engine wasn't inited.\n" );
        return FALSE;
    }
    if( Timer == INVALID_TIMER )
    {
        dprintLib( "Invalid timer requested.\n" );
        return FALSE;
    }
    else if( Timer == INFINITE_TIMER )
    {
        return TRUE;
    }
    /* It's guard *ONLY* against on thread loop:
     *  alarm -> lock(); call -> user_proc() -> and user call us recursively lock()
     */
    if( SelfLocking() ) 
    {
        return FALSE;
    }    
    if( !LockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't lock timer engine.\n" );
        return FALSE;
    }
    TAILQ_FOREACH( tmpTimer, &TimerQueue.all_head, all )
    {
        if( tmpTimer == Timer )
        {
            found = TRUE;
            break;
        }
    }
    if( !found )
    {   
        if( !UnLockQueue( &TimerQueue ) )
        {
            dprintLib( "Can't unlock timer engine.\n" );
        }
        dprintLib( "Invalid timer for clearing: not exist.\n" );
        dprintLib( "Requested timer at %p.\n", Timer );
        return FALSE;
    }
    RemoveTimer( Timer, TRUE );
    if( !UnLockQueue( &TimerQueue ) )
    {
        dprintLib( "Can't unlock timer engine.\n" );
        return FALSE;
    }
    return TRUE;
}

#endif /* __DWC_TIMO_C */
