#ifndef __DWC_UNIX_C
#define __DWC_UNIX_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_unix.c,v 1.10 2001/06/14 06:06:34 cvser Exp $";
#endif

#if defined(__UNIX_DW)

#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include "dwc_lib.h"
#include "dwc_unix.h"
#include "dwc_err.h"
#include "dwc_macro.h"

static int unix_error;

/* -------------------------------------------------------------------- */
DWC_error_code_t init_unix_sockets ( void )
{
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dinit_unix_sockets ( void )
{
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_string_t last_unix_error ( void )
{
    return strerror( unix_error );
}

/* -------------------------------------------------------------------- */
DWC_error_code_t addr_unix_op ( DWC_cmd_t cmd, DWC_unix_socket_addr_t* addr, DWC_string_t name )
{
    DWC_bool_t           used = FALSE;
    DWC_error_code_t     error;
    
    PARAM_ASSERT( addr );
    PARAM_ASSERT( name );
    
    if( cmd == DWC_IPC_CMD_SETADDR )
    {
        used = TRUE;
        if( name[0] == '\0' )
        {
            error = DWC_BADPARAMS;
            goto error_exit;
        }
        strncpy( addr->sun_path, name, sizeof(addr->sun_path) );
        addr->sun_path[sizeof(addr->sun_path)-1] = NULLCHAR;
    }
    if( !used )
    {
        error = DWC_BADPARAMS;
        goto error_exit;
    }
    error = DWC_NOERROR;
  error_exit:
    return error;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t create_unix_socket( DWC_unix_socket_t *sock, DWC_bool_t nonblocking )
{
    int                     opt;
    
    PARAM_ASSERT( sock );
    
    *sock = socket( PF_UNIX, SOCK_STREAM, 0 );
    if( (*sock) == INVALID_UNIX_SOCKET )
    {
        dprintLib( "Can't socket() on unix socket.\n" );
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    if( nonblocking )
    {
        opt = fcntl( (*sock), F_GETFL, 0 );
        if( fcntl( (*sock), F_SETFL, opt | O_NONBLOCK ) == -1 )
        {
            dprintLib("Can't set nonblocking mode for UNIX socket.\n");
            close( (*sock) );
            *sock = INVALID_UNIX_SOCKET;
            return DWC_IPCSYSTEM;
        }
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t bind_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t* addr )
{
    DWC_socklen_t len = sizeof(DWC_unix_socket_addr_t);
    
    if( bind( sock, (struct sockaddr*) addr, len ) < 0 )
    {
        dprintLib("Can't bind() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t listen_unix_socket ( DWC_unix_socket_t sock, int backlog )
{
    if( listen( sock, backlog ) != 0 )
    {
        dprintLib("Can't listen() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t accept_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t *addr, DWC_unix_socket_t *newsock )
{
    DWC_socklen_t len = sizeof(DWC_unix_socket_addr_t);
    
    *newsock = accept( sock, (struct sockaddr*)  addr, &len );
    if( (*newsock) == INVALID_UNIX_SOCKET )
    {
        dprintLib("Can't accept() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t connect_unix_socket ( DWC_unix_socket_t sock, DWC_unix_socket_addr_t *addr, DWC_timeout_t timeout )
{
    DWC_error_code_t  err;
    int               how, retval;

    retval = connect( sock, (struct sockaddr*) addr, sizeof(DWC_unix_socket_addr_t) );
    if( retval < 0 )
    {
        if( retval == EINPROGRESS )
        {
            how = DWC_IPC_WAIT_EXCP | DWC_IPC_WAIT_SEND;
            err = wait_unix_socket( sock, &how, timeout );
            if( err == DWC_NOERROR  )
            {
                if( how & DWC_IPC_WAIT_SEND )
                {
                    return DWC_NOERROR;
                }
                dprintLib("Connection failed.\n");
                err = DWC_IPCSYSTEM;
            }
            return err;
        }
        dprintLib("Can't connect() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t send_unix_socket ( DWC_unix_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, DWC_timeout_t timeout )
{
    int sended;
    
    sended = send( sock, buf, *len, 0 );
    if( sended <= 0 )
    {
        (*len) = 0;
		if( sended == EWOULDBLOCK )
		{
			return DWC_NOERROR;
		}
        dprintLib("Can't send() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    (*len) = sended;
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t recv_unix_socket ( DWC_unix_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, DWC_timeout_t timeout )
{
    int readed;
    
    readed = recv( sock, buf, *len, 0 );
    if( readed <= 0 )
    {
        (*len) = 0;
		if( readed == EWOULDBLOCK )
		{
			return DWC_NOERROR;
		}
        dprintLib("Can't recv() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    (*len) = readed;
    return DWC_NOERROR;
}
/* -------------------------------------------------------------------- */
DWC_error_code_t wait_unix_socket( DWC_unix_socket_t sock, int *how, DWC_timeout_t timeout )
{
    fd_set         rd_set, wr_set, ex_set;    
    struct timeval timo, *pt;
    int            rd;
    
    FD_ZERO( &rd_set );
    FD_ZERO( &wr_set );
    FD_ZERO( &ex_set );
    if( (*how & DWC_IPC_WAIT_RECV) == DWC_IPC_WAIT_RECV )
    {
        FD_SET( sock, &rd_set );
    }
    if( (*how & DWC_IPC_WAIT_SEND) == DWC_IPC_WAIT_SEND )
    {
        FD_SET( sock, &wr_set );
    }
    if( (*how & DWC_IPC_WAIT_EXCP) == DWC_IPC_WAIT_EXCP )
    {
        FD_SET( sock, &ex_set );
    }
    *how = 0;
    pt = NULL;
    if( timeout > 0 )
    {
        timo.tv_sec = timeout / 1000000;
        timo.tv_usec = timeout % 1000000;
        pt = &timo;
    }
    rd = select( sock+1, &rd_set, &wr_set, &ex_set, pt );
    if( (rd == 0) && pt )
    {
        return DWC_TIMED;
    }
    if( rd < 0 )
    {
        dprintLib("Error select() on TCP socket.\n");
        return DWC_IPCSYSTEM;
    }
    if( FD_ISSET( sock, &rd_set ) )
    {
        *how |= DWC_IPC_WAIT_RECV;
    }
    if( FD_ISSET( sock, &wr_set ) )
    {
        *how |= DWC_IPC_WAIT_SEND;
    }
    if( FD_ISSET( sock, &ex_set ) )
    {
        *how |= DWC_IPC_WAIT_EXCP;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t shutdown_unix_socket ( DWC_unix_socket_t *sock, DWC_unix_socket_addr_t addr, DWC_bool_t linked )
{
    int serror, uerror;

    serror = shutdown( *sock, SHUT_RDWR );        
    if( linked )
    {
        uerror = unlink( addr.sun_path );
    }
    else
    {
        uerror = DWC_NOERROR;
    }
    (*sock) = INVALID_UNIX_SOCKET;
    if( serror || uerror )
    {
        dprintLib("Can't shutdown() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t close_unix_socket ( DWC_unix_socket_t *sock, DWC_unix_socket_addr_t addr, DWC_bool_t linked )
{
    int cerror, uerror;
    
    cerror = close( *sock );    
    if( linked )
    {
        uerror = unlink( addr.sun_path );
    }
    else
    {
        uerror = DWC_NOERROR;
    }
    *sock = INVALID_UNIX_SOCKET;
    if( cerror || uerror )
    {
        dprintLib("Can't close() on unix socket.\n");
        unix_error = errno;
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
void cancel_unix_call( DWC_pointer_t sock )
{
#ifndef SHUT_RDWR
#define SHUT_RDWR   2
#endif /* SHUT_RDWR */
    shutdown( *((DWC_unix_socket_t*)sock), SHUT_RDWR ); 
}


#endif /* __UNIX_DW */

#endif /* __DWC_UNIX_H */
